/* Test mpm functions.

   Copyright 2001, 2002, 2007 Kevin Ryde

   This file is part of MPM.

   MPM is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2, or (at your option) any later version.

   MPM is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along
   with MPM; see the file COPYING.  If not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
   USA.  */

#include <stdio.h>
#include <stdlib.h>
#include "gmp.h"
#include "gmp-impl.h"
#include "tests.h"
#include "mpm.h"
#include "mpm-impl.h"


unsigned long  mbits;
mpm_t  m;
mpz_t  mz, m2powneg1z, xz, yz, wantz, gotz, dummyz;
mpr_t  x, y, got;
unsigned long  xu, yu, wantu, gotu;
long           xl, yl;
const char  *mtype;
gmp_randstate_ptr rands;
int    want_i, got_i;


static void
check_sprp (void)
{
  static struct {
    int want;
    const char *base;
    const char *m;
  } data[] = {
    { 1, "2", "17" },   /* prime */
    { 1, "2", "941" },  /* prime */
    { 0, "2", "961" },  /* composite */
    { 1, "2", "2047" }, /* first spsp 2 */
  };

  mpm_t m;
  mpr_t base;
  int i, got;

  for (i = 0; i < sizeof(data)/sizeof(data[0]); i++)
    {
      mpm_init_str (m, data[i].m, 0);
      mpr_init_set_str (base, data[i].base, 0, m);
      got = mpr_sprp_p (base, m);
      if (got != data[i].want)
        {
          printf    ("mpr_sprp_p wrong\n");
          mpz_trace ("  m    ", mpm_modref(m));
          mpr_trace ("  base ", base, m);
          printf    ("  got  %d\n", got);
          printf    ("  want %d\n", data[i].want);
          abort ();
        }

      mpr_clear (base, m);
      mpm_clear (m);
    }
}

static void
f_mpr_neg (mpr_ptr w, mpr_srcptr x, mpm_srcptr m)
{
  mpr_neg (w, x, m);
}
static void
f_mpr_add (mpr_ptr w, mpr_srcptr x, mpr_srcptr y, mpm_srcptr m)
{
  mpr_add (w, x, y, m);
}
static void
f_mpr_mul (mpr_ptr w, mpr_srcptr x, mpr_srcptr y, mpm_srcptr m)
{
  mpr_mul (w, x, y, m);
}
static void
f_mpr_pow (mpr_ptr w, mpr_srcptr x, mpz_srcptr y, mpm_srcptr m)
{
  mpr_pow (w, x, y, m);
}
static void
f_mpr_sub (mpr_ptr w, mpr_srcptr x, mpr_srcptr y, mpm_srcptr m)
{
  mpr_sub (w, x, y, m);
}


static void
check_mod (void)
{
  gmp_randstate_ptr rands = RANDS;
  mpz_srcptr  mz = MZ(m);
  mpr_t  wantr, gotr, xr, yr;
  int    xyi, t;
  int    error = 0;

  mpr_init (xr, m);
  mpr_init (yr, m);
  mpr_init (wantr, m);
  mpr_init (gotr, m);

  for (xyi = 0; xyi < 200; xyi++)
    {
      mpz_errandomb_nonzero (xz, rands, mbits);
      mpz_errandomb_nonzero (yz, rands, mbits);
      mpr_set_z (xr, xz, m);
      mpr_set_z (yr, yz, m);

      /* Unary */
      {
        static const struct {
          const char *name;
          void (*rfun) __GMP_PROTO ((mpr_ptr, mpr_srcptr, mpm_srcptr));
          void (*zfun) __GMP_PROTO ((mpz_ptr, mpz_srcptr));
        } table[] = {
          { "neg", f_mpr_neg, mpz_neg },
        };

        for (t = 0; t < numberof (table); t++)
          {
            (*table[t].zfun) (wantz, xz);
            mpz_fdiv_r (wantz, wantz, mz);
            mpr_set_z (wantr, wantz, m);

            (*table[t].rfun) (gotr, xr, m);
            mpr_get_z (gotz, gotr, m);

            if (! mpr_equal (gotr, wantr, m))
              {
                printf ("%s: not mpr_equal\n", table[t].name);
                error = 1;
              }
            if (mpz_cmp (gotz, wantz) != 0)
              {
                printf ("%s: mpz_cmp != 0\n", table[t].name);
                error = 1;
              }
            if (error)
              {
                printf     ("  xyi=%d\n", xyi);
                gmp_printf ("  m    = 0x%ZX\n", mz);
                gmp_printf ("  x    = 0x%ZX\n", xz);
                gmp_printf ("  want = 0x%ZX\n", wantz);
                gmp_printf ("  got  = 0x%ZX\n", gotz);
                printf     ("  x    "); mpr_dump (xr, m);
                printf     ("  want "); mpr_dump (wantr, m);
                printf     ("  got  "); mpr_dump (gotr, m);
                printf     ("  m    "); mpm_dump (m);
                abort ();
              }
          }
      }

      /* Binary */
      {
        static const struct {
          const char *name;
          void (*rfun) __GMP_PROTO ((mpr_ptr, mpr_srcptr, mpr_srcptr,
                                     mpm_srcptr));
          void (*zfun) __GMP_PROTO ((mpz_ptr, mpz_srcptr, mpz_srcptr));
        } table[] = {
          { "add", f_mpr_add, mpz_add },
          { "sub", f_mpr_sub, mpz_sub },
          { "mul", f_mpr_mul, mpz_mul },
        };

        for (t = 0; t < numberof (table); t++)
          {
            (*table[t].zfun) (wantz, xz, yz);
            mpz_fdiv_r (wantz, wantz, mz);
            mpr_set_z (wantr, wantz, m);

            (*table[t].rfun) (gotr, xr, yr, m);
            mpr_get_z (gotz, gotr, m);

            if (! mpr_equal (gotr, wantr, m))
              {
                printf ("%s: not mpr_equal\n", table[t].name);
                error = 1;
              }
            if (mpz_cmp (gotz, wantz) != 0)
              {
                printf ("%s: mpz_cmp != 0\n", table[t].name);
                error = 1;
              }
            if (error)
              {
                printf     ("  xyi=%d\n", xyi);
                gmp_printf ("  m    = 0x%ZX\n", mz);
                gmp_printf ("  x    = 0x%ZX\n", xz);
                gmp_printf ("  y    = 0x%ZX\n", yz);
                gmp_printf ("  want = 0x%ZX\n", wantz);
                gmp_printf ("  got  = 0x%ZX\n", gotz);
                printf     ("  x    "); mpr_dump (xr, m);
                printf     ("  y    "); mpr_dump (yr, m);
                printf     ("  want "); mpr_dump (wantr, m);
                printf     ("  got  "); mpr_dump (gotr, m);
                printf     ("  m    "); mpm_dump (m);
                abort ();
              }
          }
      }

      /* Binary with mpz */
      {
        static const struct {
          const char *name;
          void (*rfun) __GMP_PROTO ((mpr_ptr, mpr_srcptr, mpz_srcptr,
                                     mpm_srcptr));
          void (*zfun) __GMP_PROTO ((mpz_ptr, mpz_srcptr, mpz_srcptr,
                                     mpz_srcptr));
        } table[] = {
          { "pow", f_mpr_pow, mpz_powm },
        };

        for (t = 0; t < numberof (table); t++)
          {
            (*table[t].zfun) (wantz, xz, yz, mz);
            mpr_set_z (wantr, wantz, m);

            (*table[t].rfun) (gotr, xr, yz, m);
            mpr_get_z (gotz, gotr, m);

            if (! mpr_equal (gotr, wantr, m))
              {
                printf ("%s: not mpr_equal\n", table[t].name);
                error = 1;
              }
            if (mpz_cmp (gotz, wantz) != 0)
              {
                printf ("%s: mpz_cmp != 0\n", table[t].name);
                error = 1;
              }
            if (error)
              {
                printf     ("  xyi=%d\n", xyi);
                gmp_printf ("  m    = 0x%ZX\n", mz);
                gmp_printf ("  x    = 0x%ZX\n", xz);
                gmp_printf ("  y    = 0x%ZX\n", yz);
                gmp_printf ("  want = 0x%ZX\n", wantz);
                gmp_printf ("  got  = 0x%ZX\n", gotz);
                printf     ("  x    "); mpr_dump (xr, m);
                printf     ("  y    "); mpr_dump (yr, m);
                printf     ("  want "); mpr_dump (wantr, m);
                printf     ("  got  "); mpr_dump (gotr, m);
                printf     ("  m    "); mpm_dump (m);
                abort ();
              }
          }
      }
    }

  mpr_clear (xr, m);
  mpr_clear (yr, m);
  mpr_clear (wantr, m);
  mpr_clear (gotr, m);
}

static void
error_unary_ui (const char *name)
{
  printf     ("mpr_%s wrong\n", name);
  printf     ("  mtype = %s\n", mtype);
  gmp_printf ("  mz    = %#ZX\n", mz);
  printf     ("  xu    = 0x%lX\n", xu);
  gmp_printf ("  wantz = %#ZX\n", wantz);
  gmp_printf ("  gotz  = %#ZX\n", gotz);
  printf     ("  got  "); mpr_dump (got, m);
  printf     ("  m    "); mpm_dump (m);
  abort ();
}

static void
error_unary_si (const char *name)
{
  printf     ("mpr_%s wrong\n", name);
  printf     ("  mtype = %s\n", mtype);
  gmp_printf ("  mz    = %#ZX\n", mz);
  printf     ("  xl    = %ld (0x%lX)\n", xl, xl);
  gmp_printf ("  wantz = %#ZX\n", wantz);
  gmp_printf ("  gotz  = %#ZX\n", gotz);
  printf     ("  got  "); mpr_dump (got, m);
  printf     ("  m    "); mpm_dump (m);
  abort ();
}

static void
error_unary_z (const char *name)
{
  printf     ("mpr_%s wrong\n", name);
  printf     ("  mtype = %s\n", mtype);
  gmp_printf ("  mz    = %#ZX\n", mz);
  gmp_printf ("  x     = %#ZX\n", xz);
  gmp_printf ("  wantz = %#ZX\n", wantz);
  gmp_printf ("  gotz  = %#ZX\n", gotz);
  printf     ("  got  "); mpr_dump (got, m);
  printf     ("  m    "); mpm_dump (m);
  abort ();
}

static void
error_binary_si (const char *name)
{
  printf     ("mpr_%s wrong\n", name);
  printf     ("  mtype = %s\n", mtype);
  gmp_printf ("  mz    = %#ZX\n", mz);
  gmp_printf ("  xz    = %#ZX\n", xz);
  printf     ("  yl    = %ld 0x%lX\n", yl, yl);
  gmp_printf ("  wantz = %#ZX\n", wantz);
  gmp_printf ("  gotz  = %#ZX\n", gotz);
  printf     ("  x    "); mpr_dump (x, m);
  printf     ("  got  "); mpr_dump (got, m);
  printf     ("  m    "); mpm_dump (m);
  abort ();
}

static void
error_get_ui (void)
{
  printf     ("mpr_get_ui wrong\n");
  printf     ("  mtype = %s\n", mtype);
  gmp_printf ("  mz    = %#ZX\n", mz);
  gmp_printf ("  xz    = %#ZX\n", xz);
  gmp_printf ("  wantz = %#ZX\n", wantz);
  printf     ("  wantu = %#lX\n", wantu);
  gmp_printf ("  gotu  = %#lX\n", gotu);
  printf     ("  x    "); mpr_dump (x, m);
  printf     ("  m    "); mpm_dump (m);
  abort ();
}

static void
error_i_unary (const char *name)
{
  printf     ("mpr_%s wrong\n", name);
  printf     ("  mtype  = %s\n", mtype);
  gmp_printf ("  mz     = %#ZX\n", mz);
  gmp_printf ("  xz     = %#ZX\n", xz);
  printf     ("  want_i = %d\n", want_i);
  printf     ("  got_i  = %d\n", got_i);
  printf     ("  x    "); mpr_dump (x, m);
  printf     ("  m    "); mpm_dump (m);
  abort ();
}

static void
error_equal (void)
{
  printf     ("mpr_equal wrong\n");
  printf     ("  mtype = %s\n", mtype);
  gmp_printf ("  mz    = %#ZX\n", mz);
  gmp_printf ("  xz    = %#ZX\n", xz);
  gmp_printf ("  yz    = %#ZX\n", yz);
  printf     ("  x    "); mpr_dump (x, m);
  printf     ("  y    "); mpr_dump (y, m);
  printf     ("  want_i = %d\n", want_i);
  printf     ("  got_i  = %d\n", got_i);
  printf     ("  m    "); mpm_dump (m);
  abort ();
}

static void
error_equal_val (const char *val)
{
  printf     ("mpr_equal_%s wrong\n", val);
  printf     ("  mtype = %s\n", mtype);
  gmp_printf ("  mz    = %#ZX\n", mz);
  gmp_printf ("  xz    = %#ZX\n", xz);
  printf     ("  x    "); mpr_dump (x, m);
  printf     ("  want_i = %d\n", want_i);
  printf     ("  got_i  = %d\n", got_i);
  printf     ("  m    "); mpm_dump (m);
  abort ();
}

static void
error_invert (void)
{
  printf     ("mpr_invert wrong\n");
  printf     ("  mtype  = %s\n", mtype);
  gmp_printf ("  mz     = %#ZX\n", mz);
  gmp_printf ("  xz     = %#ZX\n", xz);
  gmp_printf ("  wantz  = %#ZX\n", wantz);
  gmp_printf ("  gotz   = %#ZX\n", gotz);
  printf     ("  want_i = %d\n", want_i);
  printf     ("  got_i  = %d\n", got_i);
  mpr_trace  ("  x      ", x, m);
  mpr_trace  ("  got    ", got, m);
  printf     ("  m      "); mpm_dump (m);
  abort ();
}

#define CHECK(name, error)              \
  do {                                  \
    MPZ_CHECK_FORMAT (gotz);            \
    if (mpz_cmp (gotz, wantz) != 0)     \
      error (name);                     \
  } while (0)

#define CHECK_INVERT()                                  \
  do {                                                  \
    MPZ_CHECK_FORMAT (gotz);                            \
    if ((got_i!=0) != (want_i!=0)                       \
        || (got_i && mpz_cmp (gotz, wantz) != 0))       \
      error_invert ();                                  \
  } while (0)


static void
check_one_aorsmul_ui (const char *name)
{
  MPZ_CHECK_FORMAT (gotz);
  if (mpz_cmp (gotz, wantz) != 0)
    {
      printf     ("mpr_%s wrong\n", name);
      printf     ("  mtype = %s\n", mtype);
      gmp_printf ("  mz    = %#ZX\n", mz);
      gmp_printf ("  xz    = %#ZX\n", xz);
      gmp_printf ("  yz    = %#ZX\n", yz);
      printf     ("  yu    = 0x%lX\n", yu);
      gmp_printf ("  wantz = %#ZX\n", wantz);
      gmp_printf ("  gotz  = %#ZX\n", gotz);
      printf     ("  x    "); mpr_dump (x, m);
      printf     ("  y    "); mpr_dump (y, m);
      printf     ("  got  "); mpr_dump (got, m);
      printf     ("  m    "); mpm_dump (m);
      abort ();
    }
}

static void
check_one_aorsmul_ui_inplace (const char *name)
{
  MPZ_CHECK_FORMAT (gotz);
  if (mpz_cmp (gotz, wantz) != 0)
    {
      printf     ("mpr_%s wrong in-place\n", name);
      printf     ("  mtype = %s\n", mtype);
      gmp_printf ("  mz    = %#ZX\n", mz);
      gmp_printf ("  xz    = %#ZX\n", xz);
      printf     ("  xu    = 0x%lX\n", yu);
      gmp_printf ("  wantz = %#ZX\n", wantz);
      gmp_printf ("  gotz  = %#ZX\n", gotz);
      printf     ("  x    "); mpr_dump (x, m);
      printf     ("  got  "); mpr_dump (got, m);
      printf     ("  m    "); mpm_dump (m);
      abort ();
    }
}

static void
check_one_binary (const char *name)
{
  MPZ_CHECK_FORMAT (gotz);
  if (mpz_cmp (gotz, wantz) != 0)
    {
      printf     ("mpr_%s wrong\n", name);
      printf     ("  mtype = %s\n", mtype);
      gmp_printf ("  mz    = %#ZX\n", mz);
      gmp_printf ("  xz    = %#ZX\n", xz);
      gmp_printf ("  yz    = %#ZX\n", yz);
      gmp_printf ("  wantz = %#ZX\n", wantz);
      gmp_printf ("  gotz  = %#ZX\n", gotz);
      printf     ("  x    "); mpr_dump (x, m);
      printf     ("  y    "); mpr_dump (y, m);
      printf     ("  got  "); mpr_dump (got, m);
      printf     ("  m    "); mpm_dump (m);
      abort ();
    }
}

static void
check_one_binary_ui (const char *name)
{
  MPZ_CHECK_FORMAT (gotz);
  if (mpz_cmp (gotz, wantz) != 0)
    {
      printf     ("mpr_%s wrong\n", name);
      printf     ("  mtype = %s\n", mtype);
      gmp_printf ("  mz    = %#ZX\n", mz);
      gmp_printf ("  xz    = %#ZX\n", xz);
      printf     ("  yu    = 0x%lX\n", yu);
      gmp_printf ("  wantz = %#ZX\n", wantz);
      gmp_printf ("  gotz  = %#ZX\n", gotz);
      printf     ("  x    "); mpr_dump (x, m);
      printf     ("  got  "); mpr_dump (got, m);
      printf     ("  m    "); mpm_dump (m);
      abort ();
    }
}

static void
check_one_equal_ui (void)
{
  if ((got_i!=0) != (want_i!=0))
    {
      printf     ("mpr_equal_ui wrong\n");
      printf     ("  mtype = %s\n", mtype);
      gmp_printf ("  mz    = %#ZX\n", mz);
      gmp_printf ("  xz    = %#ZX\n", xz);
      printf     ("  yu    = %#lX\n", yu);
      printf     ("  x    "); mpr_dump (x, m);
      printf     ("  want_i = %d\n", want_i);
      printf     ("  got_i  = %d\n", got_i);
      printf     ("  m    "); mpm_dump (m);
      abort ();
    }
}

static void
check_one_unary (const char *name)
{
  MPZ_CHECK_FORMAT (gotz);
  if (mpz_cmp (gotz, wantz) != 0)
    {
      printf     ("mpr_%s wrong\n", name);
      printf     ("  mtype = %s\n", mtype);
      gmp_printf ("  mz    = %#ZX\n", mz);
      gmp_printf ("  xz    = %#ZX\n", xz);
      gmp_printf ("  wantz = %#ZX\n", wantz);
      gmp_printf ("  gotz  = %#ZX\n", gotz);
      mpr_trace  ("  x     ", x, m);
      mpr_trace  ("  got   ", got, m);
      printf     ("  m     "); mpm_dump (m);
      abort ();
    }
}


/* ------------------------------------------------------------------------- */

static void
check_samples_add_ui (void)
{
  /* 0 + 0 == 0 */
  mpz_set_ui (xz, 0L);
  yu = 0;
  mpz_set_ui (wantz, 0L);
  mpr_set_z (x, xz, m);
  mpr_add_ui (got, x, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_binary_ui ("add_ui");

  /* 0 + ULONG_MAX */
  mpz_set_ui (xz, 0L);
  yu = ULONG_MAX;
  mpz_set_ui (wantz, yu);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_add_ui (got, x, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_binary_ui ("add_ui");

  /* ULONG_MAX + ULONG_MAX */
  mpz_set_ui (xz, ULONG_MAX);
  yu = ULONG_MAX;
  mpz_add_ui (wantz, xz, yu);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_add_ui (got, x, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_binary_ui ("add_ui");

  /* -ULONG_MAX + ULONG_MAX = 0 */
  mpz_set_ui (xz, ULONG_MAX);
  mpz_neg (xz, xz);
  yu = ULONG_MAX;
  mpz_set_ui (wantz, 0L);
  mpr_set_z (x, xz, m);
  mpr_add_ui (got, x, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_binary_ui ("add_ui");
}

static void
check_samples_addmul_ui (void)
{
  /* 0 + 0*0 == 0 */
  mpz_set_ui (xz, 0L);
  mpz_set_ui (yz, 0L);
  yu = 0;
  mpz_set_ui (wantz, 0L);
  mpr_set_z (x, xz, m);
  mpr_set_z (y, yz, m);
  mpr_set (got, x, m);
  mpr_addmul_ui (got, y, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_aorsmul_ui ("addmul_ui");

  /* 1 + 1*1 == 2 */
  mpz_set_ui (xz, 1L);
  mpz_set_ui (yz, 1L);
  yu = 1;
  mpz_set_ui (wantz, 2L);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_set_z (y, yz, m);
  mpr_set (got, x, m);
  mpr_addmul_ui (got, y, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_aorsmul_ui ("addmul_ui");

  /* 1 + 1*ULONG_MAX == ULONG_MAX+1 */
  mpz_set_ui (xz, 1L);
  mpz_set_ui (yz, 1L);
  yu = ULONG_MAX;
  mpz_set_ui (wantz, ULONG_MAX);
  mpz_add_ui (wantz, wantz, 1L);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_set_z (y, yz, m);
  mpr_set (got, x, m);
  mpr_addmul_ui (got, y, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_aorsmul_ui ("addmul_ui");

  /* (m-1) + (m-1)*ULONG_MAX == -(ULONG_MAX+1) */
  mpz_sub_ui (xz, mz, 1L);
  mpz_sub_ui (yz, mz, 1L);
  yu = ULONG_MAX;
  mpz_set (wantz, xz);
  mpz_addmul_ui (wantz, yz, yu);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_set_z (y, yz, m);
  mpr_set (got, x, m);
  mpr_addmul_ui (got, y, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_aorsmul_ui ("addmul_ui");
}

static void
check_samples_addmul_ui_inplace (void)
{
  /* 1 + 1*ULONG_MAX == ULONG_MAX+1 */
  mpz_set_ui (xz, 1L);
  xu = ULONG_MAX;
  mpz_set (wantz, xz);
  mpz_addmul_ui (wantz, wantz, xu);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_set (got, x, m);
  mpr_addmul_ui (got, got, xu, m);
  mpr_get_z (gotz, got, m);
  check_one_aorsmul_ui_inplace ("addmul_ui");

  /* (m-1) + (m-1)*ULONG_MAX == -(ULONG_MAX+1) */
  mpz_sub_ui (xz, mz, 1L);
  xu = ULONG_MAX;
  mpz_set (wantz, xz);
  mpz_addmul_ui (wantz, wantz, xu);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_set (got, x, m);
  mpr_addmul_ui (got, got, xu, m);
  mpr_get_z (gotz, got, m);
  check_one_aorsmul_ui_inplace ("addmul_ui");
}

static void
check_div_2exp (void)
{
  if (! mpz_odd_p (mz))
    return;

  /* 1/1 = 1 */
  mpz_set_ui (xz, 1L);
  yu = 0;
  mpz_set_ui (wantz, 1L);
  mpr_set_z (x, xz, m);
  mpr_div_2exp (got, x, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_binary_ui ("div_2exp");

  /* 8/2 = 4 */
  mpz_set_ui (xz, 8L);
  yu = 1;
  mpz_set_ui (wantz, 4L);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_div_2exp (got, x, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_binary_ui ("div_2exp");
}

static void
check_equal (void)
{
  /* 0 cf 0 */
  mpz_set_ui (xz, 0L);
  mpz_set_ui (yz, 0L);
  want_i = mpz_congruent_p (xz, yz, mz);
  mpr_set_z (x, xz, m);
  mpr_set_z (y, yz, m);
  got_i = mpr_equal (x, y, m);
  if ((got_i!=0) != (want_i!=0))
    error_equal ();

  /* 1 cf 1 */
  mpz_set_ui (xz, 1L);
  mpz_set_ui (yz, 1L);
  want_i = mpz_congruent_p (xz, yz, mz);
  mpr_set_z (x, xz, m);
  mpr_set_z (y, yz, m);
  got_i = mpr_equal (x, y, m);
  if ((got_i!=0) != (want_i!=0))
    error_equal ();

  /* 1 cf -1 */
  mpz_set_ui (xz, 1L);
  mpz_set_si (yz, -1L);
  want_i = mpz_congruent_p (xz, yz, mz);
  mpr_set_z (x, xz, m);
  mpr_set_z (y, yz, m);
  got_i = mpr_equal (x, y, m);
  if ((got_i!=0) != (want_i!=0))
    error_equal ();

  /* 0 cf ULONG_MAX */
  mpz_set_ui (xz, 0L);
  mpz_set_ui (yz, ULONG_MAX);
  want_i = mpz_congruent_p (xz, yz, mz);
  mpr_set_z (x, xz, m);
  mpr_set_z (y, yz, m);
  got_i = mpr_equal (x, y, m);
  if ((got_i!=0) != (want_i!=0))
    error_equal ();

  /* ULONG_MAX cf -ULONG_MAX */
  mpz_set_ui (xz, ULONG_MAX);
  mpz_set_ui (yz, ULONG_MAX);
  mpz_neg (yz, yz);
  want_i = mpz_congruent_p (xz, yz, mz);
  mpr_set_z (x, xz, m);
  mpr_set_z (y, yz, m);
  got_i = mpr_equal (x, y, m);
  if ((got_i!=0) != (want_i!=0))
    error_equal ();

}

static void
check_equal_negone (void)
{
  mpz_set_si (yz, -1L);

  /* 0 */
  mpz_set_ui (xz, 0L);
  mpr_set_z (x, xz, m);
  want_i = mpz_congruent_p (xz, yz, mz);
  got_i = mpr_equal_negone (x, m);
  if ((got_i!=0) != (want_i!=0))
    error_equal_val ("negone");

  /* 1 */
  mpz_set_ui (xz, 1L);
  mpr_set_z (x, xz, m);
  want_i = mpz_congruent_p (xz, yz, mz);
  got_i = mpr_equal_negone (x, m);
  if ((got_i!=0) != (want_i!=0))
    error_equal_val ("negone");

  /* 2 */
  mpz_set_ui (xz, 2L);
  mpr_set_z (x, xz, m);
  want_i = mpz_congruent_p (xz, yz, mz);
  got_i = mpr_equal_negone (x, m);
  if ((got_i!=0) != (want_i!=0))
    error_equal_val ("negone");

  /* -1 */
  mpz_set_si (xz, -1L);
  mpr_set_z (x, xz, m);
  want_i = mpz_congruent_p (xz, yz, mz);
  got_i = mpr_equal_negone (x, m);
  if ((got_i!=0) != (want_i!=0))
    error_equal_val ("negone");

  /* -2 */
  mpz_set_si (xz, -2L);
  mpr_set_z (x, xz, m);
  want_i = mpz_congruent_p (xz, yz, mz);
  got_i = mpr_equal_negone (x, m);
  if ((got_i!=0) != (want_i!=0))
    error_equal_val ("negone");
}

static void
check_equal_one (void)
{
  mpz_set_ui (yz, 1L);

  /* 0 */
  mpz_set_ui (xz, 0L);
  want_i = mpz_congruent_p (xz, yz, mz);
  mpr_set_z (x, xz, m);
  got_i = mpr_equal_one (x, m);
  if ((got_i!=0) != (want_i!=0))
    error_equal_val ("one");

  /* 1 */
  mpz_set_ui (xz, 1L);
  want_i = mpz_congruent_p (xz, yz, mz);
  mpr_set_z (x, xz, m);
  got_i = mpr_equal_one (x, m);
  if ((got_i!=0) != (want_i!=0))
    error_equal_val ("one");

  /* 2 */
  mpz_set_ui (xz, 2L);
  want_i = mpz_congruent_p (xz, yz, mz);
  mpr_set_z (x, xz, m);
  got_i = mpr_equal_one (x, m);
  if ((got_i!=0) != (want_i!=0))
    error_equal_val ("one");

  /* -1 */
  mpz_set_si (xz, -1L);
  want_i = mpz_congruent_p (xz, yz, mz);
  mpr_set_z (x, xz, m);
  got_i = mpr_equal_one (x, m);
  if ((got_i!=0) != (want_i!=0))
    error_equal_val ("one");

  /* -2 */
  mpz_set_si (xz, -2L);
  want_i = mpz_congruent_p (xz, yz, mz);
  mpr_set_z (x, xz, m);
  got_i = mpr_equal_one (x, m);
  if ((got_i!=0) != (want_i!=0))
    error_equal_val ("one");
}

static void
check_samples_equal_ui (void)
{
  /* 0 cf 0 */
  mpz_set_ui (xz, 0L);
  yu = 0;
  want_i = 1;
  mpr_set_z (x, xz, m);
  got_i = mpr_equal_ui (x, yu, m);
  check_one_equal_ui ();

  /* 1 cf 1 */
  mpz_set_ui (xz, 1L);
  yu = 1;
  want_i = 1;
  mpr_set_z (x, xz, m);
  got_i = mpr_equal_ui (x, yu, m);
  check_one_equal_ui ();

  /* 0 cf 1 */
  mpz_set_ui (xz, 0L);
  yu = 1;
  want_i = 0;
  mpr_set_z (x, xz, m);
  got_i = mpr_equal_ui (x, yu, m);
  check_one_equal_ui ();

  /* 1 cf 0 */
  mpz_set_ui (xz, 1L);
  yu = 0;
  want_i = 0;
  mpr_set_z (x, xz, m);
  got_i = mpr_equal_ui (x, yu, m);
  check_one_equal_ui ();

  /* ULONG_MAX cf ULONG_MAX */
  mpz_set_ui (xz, ULONG_MAX);
  yu = ULONG_MAX;
  want_i = 1;
  mpr_set_z (x, xz, m);
  got_i = mpr_equal_ui (x, yu, m);
  check_one_equal_ui ();

  /* 0 cf ULONG_MAX */
  mpz_set_ui (xz, 0);
  yu = ULONG_MAX;
  mpz_set_ui (yz, yu);
  want_i = mpz_congruent_p (xz, yz, mz);
  mpr_set_z (x, xz, m);
  got_i = mpr_equal_ui (x, yu, m);
  check_one_equal_ui ();

  /* 0 cf ULONG_MAX-1 */
  mpz_set_ui (xz, 0);
  yu = ULONG_MAX-1;
  mpz_set_ui (yz, yu);
  want_i = mpz_congruent_p (xz, yz, mz);
  mpr_set_z (x, xz, m);
  got_i = mpr_equal_ui (x, yu, m);
  check_one_equal_ui ();
}

static void
check_equal_zero (void)
{
  mpz_set_ui (yz, 0L);

  /* 0 */
  mpz_set_ui (xz, 0L);
  want_i = mpz_congruent_p (xz, yz, mz);
  mpr_set_z (x, xz, m);
  got_i = mpr_equal_zero (x, m);
  if ((got_i!=0) != (want_i!=0))
    error_equal_val ("zero");

  /* 1 */
  mpz_set_ui (xz, 1L);
  want_i = mpz_congruent_p (xz, yz, mz);
  mpr_set_z (x, xz, m);
  got_i = mpr_equal_zero (x, m);
  if ((got_i!=0) != (want_i!=0))
    error_equal_val ("zero");

  /* 2 */
  mpz_set_ui (xz, 2L);
  want_i = mpz_congruent_p (xz, yz, mz);
  mpr_set_z (x, xz, m);
  got_i = mpr_equal_zero (x, m);
  if ((got_i!=0) != (want_i!=0))
    error_equal_val ("zero");

  /* -1 */
  mpz_set_si (xz, -1L);
  want_i = mpz_congruent_p (xz, yz, mz);
  mpr_set_z (x, xz, m);
  got_i = mpr_equal_zero (x, m);
  if ((got_i!=0) != (want_i!=0))
    error_equal_val ("zero");
}

static void
check_invert (void)
{
  /* 1/1 = 1 */
  mpz_set_ui (xz, 1L);
  mpz_set_ui (wantz, 1L);
  want_i = 1;
  mpr_set_z (x, xz, m);
  got_i = mpr_invert (got, x, m);
  mpr_get_z (gotz, got, m);
  CHECK_INVERT ();

  /* 1/-1 = -1 */
  mpz_set_si (xz, -1L);
  mpz_sub_ui (wantz, mz, 1L);
  want_i = 1;
  mpr_set_z (x, xz, m);
  got_i = mpr_invert (got, x, m);
  mpr_get_z (gotz, got, m);
  CHECK_INVERT ();

  /* 1/2 = (m+1)/2, for m odd */
  if (mpz_odd_p (mz))
    {
      mpz_set_ui (xz, 2L);
      mpz_cdiv_q_2exp (wantz, mz, 1L);
      want_i = 1;
      mpr_set_z (x, xz, m);
      got_i = mpr_invert (got, x, m);
      mpr_get_z (gotz, got, m);
      CHECK_INVERT ();
    }

#if 0
  /* highbit */
  mpz_set_ui (xz, 0L);
  mpz_setbit (xz, mbits-1);
  mpz_invert (wantz, xz);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_invert (got, x, m);
  mpr_get_z (gotz, got, m);
  CHECK ("invert", error_unary);
#endif
}

static void
check_mul_2exp (void)
{
  /* 1 << 0 */
  mpz_set_ui (xz, 1L);
  yu = 0;
  mpz_set_ui (wantz, 1L);
  mpr_set_z (x, xz, m);
  mpr_mul_2exp (got, x, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_binary_ui ("mul_2exp");

  /* 1 << 1 */
  mpz_set_ui (xz, 1L);
  yu = 1;
  mpz_set_ui (wantz, 2L);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_mul_2exp (got, x, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_binary_ui ("mul_2exp");

  /* 123 << GMP_NUMB_BITS */
  mpz_set_ui (xz, 123L);
  yu = GMP_NUMB_BITS;
  mpz_mul_2exp (wantz, xz, yu);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_mul_2exp (got, x, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_binary_ui ("mul_2exp");

  /* 123 << 2*GMP_NUMB_BITS) */
  mpz_set_ui (xz, 123L);
  yu = 2*GMP_NUMB_BITS;
  mpz_mul_2exp (wantz, xz, yu);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_mul_2exp (got, x, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_binary_ui ("mul_2exp");

  /* (m-1) << GMP_NUMB_BITS-1 */
  mpz_set_si (xz, -1L);
  yu = GMP_NUMB_BITS-1;
  mpz_mul_2exp (wantz, xz, yu);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_mul_2exp (got, x, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_binary_ui ("mul_2exp");

  /* (m-1) << GMP_NUMB_BITS */
  mpz_set_si (xz, -1L);
  yu = GMP_NUMB_BITS;
  mpz_mul_2exp (wantz, xz, yu);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_mul_2exp (got, x, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_binary_ui ("mul_2exp");

  /* (m-1) << GMP_NUMB_BITS+1 */
  mpz_set_si (xz, -1L);
  yu = GMP_NUMB_BITS+1;
  mpz_mul_2exp (wantz, xz, yu);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_mul_2exp (got, x, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_binary_ui ("mul_2exp");
}

static void
check_mul_si (void)
{
  /* m-1 * 1 */
  mpz_sub_ui (xz, mz, 1L);
  yl = 1L;
  mpz_mul_si (wantz, xz, yl);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_mul_si (got, x, yl, m);
  mpr_get_z (gotz, got, m);
  CHECK ("mul_si", error_binary_si);

  /* m-1 * LONG_MAX-1 */
  mpz_sub_ui (xz, mz, 1L);
  yl = LONG_MAX-1;
  mpz_mul_si (wantz, xz, yl);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_mul_si (got, x, yl, m);
  mpr_get_z (gotz, got, m);
  CHECK ("mul_si", error_binary_si);

  /* m-1 * LONG_MAX */
  mpz_sub_ui (xz, mz, 1L);
  yl = LONG_MAX;
  mpz_mul_si (wantz, xz, yl);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_mul_si (got, x, yl, m);
  mpr_get_z (gotz, got, m);
  CHECK ("mul_si", error_binary_si);

  /* m-1 * LONG_MIN+1 */
  mpz_sub_ui (xz, mz, 1L);
  yl = LONG_MIN+1;
  mpz_mul_si (wantz, xz, yl);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_mul_si (got, x, yl, m);
  mpr_get_z (gotz, got, m);
  CHECK ("mul_si", error_binary_si);

  /* m-1 * LONG_MIN */
  mpz_sub_ui (xz, mz, 1L);
  yl = LONG_MIN;
  mpz_mul_si (wantz, xz, yl);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_mul_si (got, x, yl, m);
  mpr_get_z (gotz, got, m);
  CHECK ("mul_si", error_binary_si);
}

static void
check_mul_ui (void)
{
  /* m-1 * 1 */
  mpz_sub_ui (xz, mz, 1L);
  yu = 1L;
  mpz_mul_ui (wantz, xz, yu);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_mul_ui (got, x, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_binary_ui ("mul_ui");

  /* m-1 * 2 */
  mpz_sub_ui (xz, mz, 1L);
  yu = 2L;
  mpz_mul_ui (wantz, xz, yu);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_mul_ui (got, x, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_binary_ui ("mul_ui");

  /* m-1 * ULONG_MAX-1 */
  mpz_sub_ui (xz, mz, 1L);
  yu = ULONG_MAX-1;
  mpz_mul_ui (wantz, xz, yu);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_mul_ui (got, x, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_binary_ui ("mul_ui");

  /* m-1 * ULONG_MAX */
  mpz_sub_ui (xz, mz, 1L);
  yu = ULONG_MAX;
  mpz_mul_ui (wantz, xz, yu);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_mul_ui (got, x, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_binary_ui ("mul_ui");
}

static void
check_neg (void)
{
  /* 0 */
  mpz_set_ui (xz, 0L);
  mpz_set_ui (wantz, 0L);
  mpr_set_z (x, xz, m);
  mpr_neg (got, x, m);
  mpr_get_z (gotz, got, m);
  check_one_unary ("neg");

  /* 1 */
  mpz_set_ui (xz, 1L);
  mpz_neg (wantz, xz);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_neg (got, x, m);
  mpr_get_z (gotz, got, m);
  check_one_unary ("neg");

  /* -1 */
  mpz_set_si (xz, -1L);
  mpz_neg (wantz, xz);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_neg (got, x, m);
  mpr_get_z (gotz, got, m);
  check_one_unary ("neg");

  /* highbit */
  mpz_set_ui (xz, 0L);
  mpz_setbit (xz, mbits-1);
  mpz_neg (wantz, xz);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_neg (got, x, m);
  mpr_get_z (gotz, got, m);
  check_one_unary ("neg");

#if 0
  /* one bit */
  for (i = 0; i < 10; i++)
    {
      mpz_set_ui (xz, 0L);
      mpz_setbit (xz, urandom() % mbits);
      mpz_sub (wantz, mz, xz);
      mpr_set_z (x, xz, m);
      mpr_neg (got, x, m);
      mpr_get_z (gotz, got, m);
      if (mpz_cmp (gotz, wantz) != 0)
        goto error;
    }
#endif
}

static void
check_set_si (void)
{
  /* 0 */
  xl = 0L;
  mpz_set_si (wantz, xl);
  mpr_set_si (got, xl, m);
  mpr_get_z (gotz, got, m);
  CHECK ("set_si", error_unary_si);

  /* 1 */
  xl = 1L;
  mpz_set_si (wantz, xl);
  mpr_set_si (got, xl, m);
  mpr_get_z (gotz, got, m);
  CHECK ("set_si", error_unary_si);

  /* -1 */
  xl = -1;
  mpz_sub_ui (wantz, mz, 1L);
  mpr_set_si (got, xl, m);
  mpr_get_z (gotz, got, m);
  CHECK ("set_si", error_unary_si);

  /* LONG_MAX */
  xl = LONG_MAX;
  mpz_set_si (wantz, xl);
  mpz_mod (wantz, wantz, mz);
  mpr_set_si (got, xl, m);
  mpr_get_z (gotz, got, m);
  CHECK ("set_si", error_unary_si);

  /* LONG_MIN */
  xl = LONG_MIN;
  mpz_set_si (wantz, xl);
  mpz_mod (wantz, wantz, mz);
  mpr_set_si (got, xl, m);
  mpr_get_z (gotz, got, m);
  CHECK ("set_si", error_unary_si);
}

static void
check_set_ui (void)
{
  /* 0 */
  xu = 1L;
  mpz_set_ui (wantz, xu);
  mpr_set_ui (got, xu, m);
  mpr_get_z (gotz, got, m);
  CHECK ("set_ui", error_unary_ui);

  /* 1 */
  xu = 1L;
  mpz_set_ui (wantz, xu);
  mpr_set_ui (got, xu, m);
  mpr_get_z (gotz, got, m);
  CHECK ("set_ui", error_unary_ui);

  /* ULONG_MAX */
  xu = ULONG_MAX;
  mpz_set_ui (wantz, xu);
  mpz_mod (wantz, wantz, mz);
  mpr_set_ui (got, xu, m);
  mpr_get_z (gotz, got, m);
  CHECK ("set_ui", error_unary_ui);
}

static void
check_set_z (void)
{
  /* 0 */
  mpz_set_ui (xz, 0L);
  mpz_set (wantz, xz);
  mpr_set_z (got, xz, m);
  mpr_get_z (gotz, got, m);
  CHECK ("set_z", error_unary_z);

  /* 1 */
  mpz_set_ui (xz, 1L);
  mpz_set (wantz, xz);
  mpr_set_z (got, xz, m);
  mpr_get_z (gotz, got, m);
  CHECK ("set_z", error_unary_z);

  /* -1 */
  mpz_set_si (xz, -1L);
  mpz_sub_ui (wantz, mz, 1L);
  mpr_set_z (got, xz, m);
  mpr_get_z (gotz, got, m);
  CHECK ("set_z", error_unary_z);
}

static void
check_samples_sub_ui (void)
{
  /* 0 - 0 == 0 */
  mpz_set_ui (xz, 0L);
  yu = 0;
  mpz_set_ui (wantz, 0L);
  mpr_set_z (x, xz, m);
  mpr_sub_ui (got, x, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_binary_ui ("sub_ui");

  /* 0 - ULONG_MAX */
  mpz_set_ui (xz, 0L);
  yu = ULONG_MAX;
  mpz_set_ui (wantz, yu);
  mpz_neg (wantz, wantz);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_sub_ui (got, x, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_binary_ui ("sub_ui");

  /* ULONG_MAX - ULONG_MAX == 0 */
  mpz_set_ui (xz, ULONG_MAX);
  yu = ULONG_MAX;
  mpz_set_ui (wantz, 0L);
  mpr_set_z (x, xz, m);
  mpr_sub_ui (got, x, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_binary_ui ("sub_ui");
}

static void
check_samples_submul_ui (void)
{
  /* 0 + 0*0 == 0 */
  mpz_set_ui (xz, 0L);
  mpz_set_ui (yz, 0L);
  yu = 0;
  mpz_set_ui (wantz, 0L);
  mpr_set_z (x, xz, m);
  mpr_set_z (y, yz, m);
  mpr_set (got, x, m);
  mpr_submul_ui (got, y, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_aorsmul_ui ("submul_ui");

  /* 1 - 1*1 == 2 */
  mpz_set_ui (xz, 1L);
  mpz_set_ui (yz, 1L);
  yu = 1;
  mpz_set_ui (wantz, 0L);
  mpr_set_z (x, xz, m);
  mpr_set_z (y, yz, m);
  mpr_set (got, x, m);
  mpr_submul_ui (got, y, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_aorsmul_ui ("submul_ui");

  /* 0 - 1*ULONG_MAX == -ULONG_MAX */
  mpz_set_ui (xz, 0L);
  mpz_set_ui (yz, 1L);
  yu = ULONG_MAX;
  mpz_set_ui (wantz, ULONG_MAX);
  mpz_neg (wantz, wantz);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_set_z (y, yz, m);
  mpr_set (got, x, m);
  mpr_submul_ui (got, y, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_aorsmul_ui ("submul_ui");

  /* 1 - 1*ULONG_MAX == -(ULONG_MAX-1) */
  mpz_set_ui (xz, 1L);
  mpz_set_ui (yz, 1L);
  yu = ULONG_MAX;
  mpz_set_ui (wantz, ULONG_MAX-1);
  mpz_neg (wantz, wantz);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_set_z (y, yz, m);
  mpr_set (got, x, m);
  mpr_submul_ui (got, y, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_aorsmul_ui ("submul_ui");

  /* (m-1) - (m-1)*ULONG_MAX == ULONG_MAX-1 */
  mpz_sub_ui (xz, mz, 1L);
  mpz_sub_ui (yz, mz, 1L);
  yu = ULONG_MAX;
  mpz_set_ui (wantz, ULONG_MAX-1);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_set_z (y, yz, m);
  mpr_set (got, x, m);
  mpr_submul_ui (got, y, yu, m);
  mpr_get_z (gotz, got, m);
  check_one_aorsmul_ui ("submul_ui");
}

static void
check_samples_submul_ui_inplace (void)
{
  /* 1 - 1*ULONG_MAX == -(ULONG_MAX-1) */
  mpz_set_ui (xz, 1L);
  xu = ULONG_MAX;
  mpz_set_ui (wantz, ULONG_MAX-1);
  mpz_neg (wantz, wantz);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_set (got, x, m);
  mpr_submul_ui (got, got, xu, m);
  mpr_get_z (gotz, got, m);
  check_one_aorsmul_ui_inplace ("submul_ui");

  /* (m-1) - (m-1)*ULONG_MAX == ULONG_MAX-1 */
  mpz_sub_ui (xz, mz, 1L);
  xu = ULONG_MAX;
  mpz_set_ui (wantz, ULONG_MAX-1);
  mpz_mod (wantz, wantz, mz);
  mpr_set_z (x, xz, m);
  mpr_set (got, x, m);
  mpr_submul_ui (got, got, xu, m);
  mpr_get_z (gotz, got, m);
  check_one_aorsmul_ui_inplace ("submul_ui");
}

static void
check_urandomm (void)
{
  gmp_randstate_ptr  rands = RANDS;
  int  i;

  for (i = 0; i < 10; i++)
    mpr_urandomm (got, rands, m);
}

static void
check_various (void)
{
  int    i, invert_ok;

  /* random */
  for (i = 0; i < 20; i++)
    {
      mpz_errandomb (xz, rands, mbits + GMP_NUMB_BITS);
      mpz_errandomb (yz, rands, mbits + GMP_NUMB_BITS);
      mpr_set_z (x, xz, m);
      mpr_set_z (y, yz, m);
      xu = urandom ();
      yu = urandom ();

      /* add */
      mpz_add (wantz, xz, yz);
      mpz_mod (wantz, wantz, mz);
      mpr_add (got, x, y, m);
      mpr_get_z (gotz, got, m);
      check_one_binary ("add");

      /* add_ui */
      mpz_add_ui (wantz, xz, yu);
      mpz_mod (wantz, wantz, mz);
      mpr_add_ui (got, x, yu, m);
      mpr_get_z (gotz, got, m);
      check_one_binary_ui ("add_ui");

      /* addmul_ui */
      mpz_set (wantz, xz);
      mpz_addmul_ui (wantz, yz, yu);
      mpz_mod (wantz, wantz, mz);
      mpr_set (got, x, m);
      mpr_addmul_ui (got, y, yu, m);
      mpr_get_z (gotz, got, m);
      check_one_aorsmul_ui ("addmul_ui");

      /* equal */
      got_i = mpr_equal (x, x, m);
      want_i = 1;
      if ((got_i!=0) != (want_i!=0))
        error_equal ();

      /* gcd */
      mpz_gcd (wantz, xz, mz);
      mpr_gcd (gotz, x, m);
      check_one_unary ("gcd");

      /* get_ui */
      mpz_mod (wantz, xz, mz);
      wantu = mpz_get_ui (wantz);
      gotu = mpr_get_ui (gotz, x, m);
      if (gotu != wantu)
        error_get_ui ();

      /* invert */
      want_i = mpz_invert (wantz, xz, mz);
      got_i = mpr_invert (got, x, m);
      mpr_get_z (gotz, got, m);
      CHECK_INVERT ();
      invert_ok = want_i;

      /* jacobi */
      if (mpz_odd_p (mz))
        {
          want_i = mpz_jacobi (xz, mz);
          got_i = mpr_jacobi (x, m);
          if (got_i != want_i)
            error_i_unary ("jacobi");
        }

      /* mul */
      mpz_mul (wantz, xz, yz);
      mpz_mod (wantz, wantz, mz);
      mpr_mul (got, x, y, m);
      mpr_get_z (gotz, got, m);
      check_one_binary ("mul");

      /* neg */
      mpz_neg (wantz, xz);
      mpz_mod (wantz, wantz, mz);
      mpr_neg (got, x, m);
      mpr_get_z (gotz, got, m);
      check_one_unary ("neg");

      /* pow */
      mpz_powm (wantz, xz, yz, mz);
      mpr_pow (got, x, yz, m);
      mpr_get_z (gotz, got, m);
      check_one_binary ("pow");

      /* pow, negative exponent */
      if (invert_ok)
        {
          mpz_neg (yz, yz);
          mpz_powm (wantz, xz, yz, mz);
          mpr_pow (got, x, yz, m);
          mpr_get_z (gotz, got, m);
          check_one_binary ("pow");
          mpz_neg (yz, yz);
        }

      /* sub */
      mpz_sub (wantz, xz, yz);
      mpz_mod (wantz, wantz, mz);
      mpr_sub (got, x, y, m);
      mpr_get_z (gotz, got, m);
      check_one_binary ("sub");

      /* sub_ui */
      mpz_sub_ui (wantz, xz, yu);
      mpz_mod (wantz, wantz, mz);
      mpr_sub_ui (got, x, yu, m);
      mpr_get_z (gotz, got, m);
      check_one_binary_ui ("sub_ui");

      /* submul_ui */
      mpz_set (wantz, xz);
      mpz_submul_ui (wantz, yz, yu);
      mpz_mod (wantz, wantz, mz);
      mpr_set (got, x, m);
      mpr_submul_ui (got, y, yu, m);
      mpr_get_z (gotz, got, m);
      check_one_aorsmul_ui ("submul_ui");


      yu %= 5*mbits;

      /* div_2exp */
      if (mpz_odd_p (mz))
        {
          mpz_powm_ui (wantz, m2powneg1z, yu, mz);
          mpz_mul (wantz, wantz, xz);
          mpz_mod (wantz, wantz, mz);
          mpr_div_2exp (got, x, yu, m);
          mpr_get_z (gotz, got, m);
          check_one_binary_ui ("div_2exp");
        }

      /* mul_2exp */
      mpz_set_ui (wantz, 2L);
      mpz_powm_ui (wantz, wantz, yu, mz);
      mpz_mul (wantz, wantz, xz);
      mpz_mod (wantz, wantz, mz);
      mpr_mul_2exp (got, x, yu, m);
      mpr_get_z (gotz, got, m);
      check_one_binary_ui ("mul_2exp");
    }
}



static void
check_mz (void)
{
  ASSERT_ALWAYS_MPM (m);

  if (mpz_cmp (mpm_modref(m), mz) != 0)
    {
      printf ("mpm_modref wrong\n");
      printf     ("  mtype = %s\n", mtype);
      gmp_printf ("  m     = 0x%ZX\n", mz);
      gmp_printf ("  got   = 0x%ZX\n", mpm_modref(m));
      abort ();
    }

  /* 2^-1 mod m */
  mpz_set_ui (m2powneg1z, 2L);
  if (mpz_invert (m2powneg1z, m2powneg1z, mz) == 0)
    mpz_set_ui (m2powneg1z, 0L);

  mpr_init (x, m);
  mpr_init (y, m);
  mpr_init (got, m);

  check_samples_add_ui ();
  check_samples_addmul_ui ();
  check_samples_addmul_ui_inplace ();
  check_div_2exp ();
  check_equal ();
  check_samples_equal_ui ();
  check_equal_negone ();
  check_equal_one ();
  check_equal_zero ();
  check_invert ();
  check_mod ();
  check_mul_2exp ();
  check_mul_si ();
  check_mul_ui ();
  check_neg ();
  check_set_z ();
  check_set_ui ();
  check_set_si ();
  check_samples_sub_ui ();
  check_samples_submul_ui ();
  check_samples_submul_ui_inplace ();
  check_urandomm ();
  check_various ();

  mpr_clear (x, m);
  mpr_clear (y, m);
  mpr_clear (got, m);
  mpm_clear (m);
}

static void
check_mbits (void)
{
  int  i;

  mtype = "2exp";
  mpz_set_ui (mz, 0L);
  mpz_setbit (mz, mbits);
  mpm_init_2exp (m, mbits);
  check_mz ();

  mtype = "2expsub1";
  mpz_set_ui (mz, 0L);
  mpz_setbit (mz, mbits);
  mpz_sub_ui (mz, mz, 1L);
  mpm_init_2expsub1 (m, mbits);
  check_mz ();

  /* random */
  for (i = 0; i < 5; i++)
    {
      mpz_errandomb (mz, rands, mbits);
      mpz_setbit (mz, mbits-1);

      mtype = "div";
      mpm_init_div (m, mz);
      check_mz ();

      mtype = "redc";
      mpz_setbit (mz, 0L); /* odd */
      mpm_init_redc (m, mz);
      check_mz ();
    }
}

static void
check_neg_2_nonzero (void)
{
  mp_limb_t  h, l;

  NEG_2_LOWNZ (h,l, GMP_NUMB_MAX,GMP_NUMB_MAX);
  ASSERT_ALWAYS (h == 0);
  ASSERT_ALWAYS (l == 1);

  NEG_2_LOWNZ (h,l, 0,1);
  ASSERT_ALWAYS (h == GMP_NUMB_MAX);
  ASSERT_ALWAYS (l == GMP_NUMB_MAX);
}

static void
check_neg_3_nonzero (void)
{
  mp_limb_t  w2,w1,w0;

  NEG_3_LOWNZ (w2,w1,w0, GMP_NUMB_MAX,GMP_NUMB_MAX,GMP_NUMB_MAX);
  ASSERT_ALWAYS (w2 == 0);
  ASSERT_ALWAYS (w1 == 0);
  ASSERT_ALWAYS (w0 == 1);

  NEG_3_LOWNZ (w2,w1,w0, 0,0,1);
  ASSERT_ALWAYS (w2 == GMP_NUMB_MAX);
  ASSERT_ALWAYS (w1 == GMP_NUMB_MAX);
  ASSERT_ALWAYS (w0 == GMP_NUMB_MAX);
}

static void
check_all (void)
{
  mpz_init (mz);
  mpz_init (m2powneg1z);
  mpz_init (xz);
  mpz_init (yz);
  mpz_init (gotz);
  mpz_init (wantz);
  mpz_init (dummyz);

  check_sprp ();
  check_neg_2_nonzero ();
  check_neg_3_nonzero ();

  for (mbits = 2; mbits < 10 * GMP_NUMB_BITS; mbits++)
    check_mbits ();

  mpz_clear (mz);
  mpz_clear (m2powneg1z);
  mpz_clear (xz);
  mpz_clear (yz);
  mpz_clear (gotz);
  mpz_clear (wantz);
  mpz_clear (dummyz);
}

int
main (void)
{
  tests_start ();
  rands = RANDS;
  mp_trace_base = 10;

  check_all ();

  tests_end ();
  return 0;
}
